@@ -10,7 +10,7 @@
    This file is part of LibVEX, a library for dynamic binary
    instrumentation and translation.
 
-   Copyright (C) 2004-2005 OpenWorks LLP.  All rights reserved.
+   Copyright (C) 2004-2007 OpenWorks LLP.  All rights reserved.
 
    This library is made available under a dual licensing scheme.
 
@@ -50,6 +50,7 @@
 #include "libvex_guest_amd64.h"
 #include "libvex_guest_arm.h"
 #include "libvex_guest_ppc32.h"
+#include "libvex_guest_ppc64.h"
 
 #include "main/vex_globals.h"
 #include "main/vex_util.h"
@@ -58,17 +59,23 @@
 
 #include "host-x86/hdefs.h"
 #include "host-amd64/hdefs.h"
-#include "host-ppc32/hdefs.h"
+#include "host-ppc/hdefs.h"
 
 #include "guest-generic/bb_to_IR.h"
 #include "guest-x86/gdefs.h"
 #include "guest-amd64/gdefs.h"
 #include "guest-arm/gdefs.h"
-#include "guest-ppc32/gdefs.h"
+#include "guest-ppc/gdefs.h"
 
 
 /* This file contains the top level interface to the library. */
 
+/* --------- fwds ... --------- */
+
+static Bool   are_valid_hwcaps ( VexArch arch, UInt hwcaps );
+static HChar* show_hwcaps ( VexArch arch, UInt hwcaps );
+
+
 /* --------- Initialise the library. --------- */
 
 /* Exported to library client. */
@@ -172,66 +179,32 @@ void LibVEX_Init (
 
 /* Exported to library client. */
 
-VexTranslateResult LibVEX_Translate (
-   /* The instruction sets we are translating from and to. */
-   VexArch      arch_guest,
-   VexArchInfo* archinfo_guest,
-   VexArch      arch_host,
-   VexArchInfo* archinfo_host,
-   /* IN: the block to translate, and its guest address. */
-   /* where are the actual bytes in the host's address space? */
-   UChar*  guest_bytes,
-   /* where do the bytes came from in the guest's aspace? */
-   Addr64  guest_bytes_addr,
-   /* what guest entry point address do they correspond to? */
-   Addr64  guest_bytes_addr_noredir,
-   /* Is it OK to chase into this guest address? */
-   Bool    (*chase_into_ok) ( Addr64 ),
-   /* OUT: which bits of guest code actually got translated */
-   VexGuestExtents* guest_extents,
-   /* IN: a place to put the resulting code, and its size */
-   UChar*  host_bytes,
-   Int     host_bytes_size,
-   /* OUT: how much of the output area is used. */
-   Int*    host_bytes_used,
-   /* IN: optionally, two instrumentation functions. */
-   IRBB*   (*instrument1) ( IRBB*, VexGuestLayout*, 
-                            Addr64, VexGuestExtents*, 
-                            IRType gWordTy, IRType hWordTy ),
-   IRBB*   (*instrument2) ( IRBB*, VexGuestLayout*, 
-                            Addr64, VexGuestExtents*,
-                            IRType gWordTy, IRType hWordTy ),
-   Bool    cleanup_after_instrumentation,
-   /* IN: should this translation be self-checking? */
-   Bool    do_self_check,
-   /* IN: optionally, an access check function for guest code. */
-   Bool    (*byte_accessible) ( Addr64 ),
-   /* IN: debug: trace vex activity at various points */
-   Int     traceflags
-)
+VexTranslateResult LibVEX_Translate ( VexTranslateArgs* vta )
 {
    /* This the bundle of functions we need to do the back-end stuff
       (insn selection, reg-alloc, assembly) whilst being insulated
       from the target instruction set. */
    HReg* available_real_regs;
    Int   n_available_real_regs;
-   Bool         (*isMove)      (HInstr*, HReg*, HReg*);
-   void         (*getRegUsage) (HRegUsage*, HInstr*);
-   void         (*mapRegs)     (HRegRemap*, HInstr*);
-   HInstr*      (*genSpill)    ( HReg, Int );
-   HInstr*      (*genReload)   ( HReg, Int );
-   void         (*ppInstr)     ( HInstr* );
-   void         (*ppReg)       ( HReg );
-   HInstrArray* (*iselBB)      ( IRBB*, VexArchInfo* );
-   Int          (*emit)        ( UChar*, Int, HInstr* );
-   IRExpr*      (*specHelper)  ( HChar*, IRExpr** );
+   Bool         (*isMove)       ( HInstr*, HReg*, HReg* );
+   void         (*getRegUsage)  ( HRegUsage*, HInstr*, Bool );
+   void         (*mapRegs)      ( HRegRemap*, HInstr*, Bool );
+   HInstr*      (*genSpill)     ( HReg, Int, Bool );
+   HInstr*      (*genReload)    ( HReg, Int, Bool );
+   HInstr*      (*directReload) ( HInstr*, HReg, Short );
+   void         (*ppInstr)      ( HInstr*, Bool );
+   void         (*ppReg)        ( HReg );
+   HInstrArray* (*iselSB)       ( IRSB*, VexArch, VexArchInfo*, 
+                                                  VexAbiInfo* );
+   Int          (*emit)         ( UChar*, Int, HInstr*, Bool, void* );
+   IRExpr*      (*specHelper)   ( HChar*, IRExpr** );
    Bool         (*preciseMemExnsFn) ( Int, Int );
 
    DisOneInstrFn disInstrFn;
 
    VexGuestLayout* guest_layout;
    Bool            host_is_bigendian = False;
-   IRBB*           irbb;
+   IRSB*           irsb;
    HInstrArray*    vcode;
    HInstrArray*    rcode;
    Int             i, j, k, out_used, guest_sizeB;
@@ -239,6 +212,7 @@ VexTranslateResult LibVEX_Translate (
    UChar           insn_bytes[32];
    IRType          guest_word_type;
    IRType          host_word_type;
+   Bool            mode64;
 
    guest_layout           = NULL;
    available_real_regs    = NULL;
@@ -248,9 +222,10 @@ VexTranslateResult LibVEX_Translate (
    mapRegs                = NULL;
    genSpill               = NULL;
    genReload              = NULL;
+   directReload           = NULL;
    ppInstr                = NULL;
    ppReg                  = NULL;
-   iselBB                 = NULL;
+   iselSB                 = NULL;
    emit                   = NULL;
    specHelper             = NULL;
    preciseMemExnsFn       = NULL;
@@ -259,8 +234,9 @@ VexTranslateResult LibVEX_Translate (
    host_word_type         = Ity_INVALID;
    offB_TISTART           = 0;
    offB_TILEN             = 0;
+   mode64                 = False;
 
-   vex_traceflags = traceflags;
+   vex_traceflags = vta->traceflags;
 
    vassert(vex_initdone);
    vexSetAllocModeTEMP_and_clear();
@@ -269,61 +245,83 @@ VexTranslateResult LibVEX_Translate (
    /* First off, check that the guest and host insn sets
       are supported. */
 
-   switch (arch_host) {
+   switch (vta->arch_host) {
 
       case VexArchX86:
+         mode64       = False;
          getAllocableRegs_X86 ( &n_available_real_regs,
                                 &available_real_regs );
-         isMove      = (Bool(*)(HInstr*,HReg*,HReg*)) isMove_X86Instr;
-         getRegUsage = (void(*)(HRegUsage*,HInstr*)) getRegUsage_X86Instr;
-         mapRegs     = (void(*)(HRegRemap*,HInstr*)) mapRegs_X86Instr;
-         genSpill    = (HInstr*(*)(HReg,Int)) genSpill_X86;
-         genReload   = (HInstr*(*)(HReg,Int)) genReload_X86;
-         ppInstr     = (void(*)(HInstr*)) ppX86Instr;
-         ppReg       = (void(*)(HReg)) ppHRegX86;
-         iselBB      = iselBB_X86;
-         emit        = (Int(*)(UChar*,Int,HInstr*)) emit_X86Instr;
+         isMove       = (Bool(*)(HInstr*,HReg*,HReg*)) isMove_X86Instr;
+         getRegUsage  = (void(*)(HRegUsage*,HInstr*, Bool)) getRegUsage_X86Instr;
+         mapRegs      = (void(*)(HRegRemap*,HInstr*, Bool)) mapRegs_X86Instr;
+         genSpill     = (HInstr*(*)(HReg,Int, Bool)) genSpill_X86;
+         genReload    = (HInstr*(*)(HReg,Int, Bool)) genReload_X86;
+         directReload = (HInstr*(*)(HInstr*,HReg,Short)) directReload_X86;
+         ppInstr      = (void(*)(HInstr*, Bool)) ppX86Instr;
+         ppReg        = (void(*)(HReg)) ppHRegX86;
+         iselSB       = iselSB_X86;
+         emit         = (Int(*)(UChar*,Int,HInstr*,Bool,void*)) emit_X86Instr;
          host_is_bigendian = False;
          host_word_type    = Ity_I32;
-         vassert(archinfo_host->subarch == VexSubArchX86_sse0
-                 || archinfo_host->subarch == VexSubArchX86_sse1
-                 || archinfo_host->subarch == VexSubArchX86_sse2);
+         vassert(are_valid_hwcaps(VexArchX86, vta->archinfo_host.hwcaps));
+         vassert(vta->dispatch != NULL); /* jump-to-dispatcher scheme */
          break;
 
       case VexArchAMD64:
+         mode64      = True;
          getAllocableRegs_AMD64 ( &n_available_real_regs,
                                   &available_real_regs );
          isMove      = (Bool(*)(HInstr*,HReg*,HReg*)) isMove_AMD64Instr;
-         getRegUsage = (void(*)(HRegUsage*,HInstr*)) getRegUsage_AMD64Instr;
-         mapRegs     = (void(*)(HRegRemap*,HInstr*)) mapRegs_AMD64Instr;
-         genSpill    = (HInstr*(*)(HReg,Int)) genSpill_AMD64;
-         genReload   = (HInstr*(*)(HReg,Int)) genReload_AMD64;
-         ppInstr     = (void(*)(HInstr*)) ppAMD64Instr;
+         getRegUsage = (void(*)(HRegUsage*,HInstr*, Bool)) getRegUsage_AMD64Instr;
+         mapRegs     = (void(*)(HRegRemap*,HInstr*, Bool)) mapRegs_AMD64Instr;
+         genSpill    = (HInstr*(*)(HReg,Int, Bool)) genSpill_AMD64;
+         genReload   = (HInstr*(*)(HReg,Int, Bool)) genReload_AMD64;
+         ppInstr     = (void(*)(HInstr*, Bool)) ppAMD64Instr;
          ppReg       = (void(*)(HReg)) ppHRegAMD64;
-         iselBB      = iselBB_AMD64;
-         emit        = (Int(*)(UChar*,Int,HInstr*)) emit_AMD64Instr;
+         iselSB      = iselSB_AMD64;
+         emit        = (Int(*)(UChar*,Int,HInstr*,Bool,void*)) emit_AMD64Instr;
          host_is_bigendian = False;
          host_word_type    = Ity_I64;
-         vassert(archinfo_host->subarch == VexSubArch_NONE);
+         vassert(are_valid_hwcaps(VexArchAMD64, vta->archinfo_host.hwcaps));
+         vassert(vta->dispatch != NULL); /* jump-to-dispatcher scheme */
          break;
 
       case VexArchPPC32:
-         getAllocableRegs_PPC32 ( &n_available_real_regs,
-                                  &available_real_regs );
-         isMove      = (Bool(*)(HInstr*,HReg*,HReg*)) isMove_PPC32Instr;
-         getRegUsage = (void(*)(HRegUsage*,HInstr*)) getRegUsage_PPC32Instr;
-         mapRegs     = (void(*)(HRegRemap*,HInstr*)) mapRegs_PPC32Instr;
-         genSpill    = (HInstr*(*)(HReg,Int)) genSpill_PPC32;
-         genReload   = (HInstr*(*)(HReg,Int)) genReload_PPC32;
-         ppInstr     = (void(*)(HInstr*)) ppPPC32Instr;
-         ppReg       = (void(*)(HReg)) ppHRegPPC32;
-         iselBB      = iselBB_PPC32;
-         emit        = (Int(*)(UChar*,Int,HInstr*)) emit_PPC32Instr;
+         mode64      = False;
+         getAllocableRegs_PPC ( &n_available_real_regs,
+                                &available_real_regs, mode64 );
+         isMove      = (Bool(*)(HInstr*,HReg*,HReg*)) isMove_PPCInstr;
+         getRegUsage = (void(*)(HRegUsage*,HInstr*,Bool)) getRegUsage_PPCInstr;
+         mapRegs     = (void(*)(HRegRemap*,HInstr*,Bool)) mapRegs_PPCInstr;
+         genSpill    = (HInstr*(*)(HReg,Int,Bool)) genSpill_PPC;
+         genReload   = (HInstr*(*)(HReg,Int,Bool)) genReload_PPC;
+         ppInstr     = (void(*)(HInstr*,Bool)) ppPPCInstr;
+         ppReg       = (void(*)(HReg)) ppHRegPPC;
+         iselSB      = iselSB_PPC;
+         emit        = (Int(*)(UChar*,Int,HInstr*,Bool,void*)) emit_PPCInstr;
          host_is_bigendian = True;
          host_word_type    = Ity_I32;
-         vassert(archinfo_guest->subarch == VexSubArchPPC32_I
-                 || archinfo_guest->subarch == VexSubArchPPC32_FI
-                 || archinfo_guest->subarch == VexSubArchPPC32_VFI);
+         vassert(are_valid_hwcaps(VexArchPPC32, vta->archinfo_host.hwcaps));
+         vassert(vta->dispatch == NULL); /* return-to-dispatcher scheme */
+         break;
+
+      case VexArchPPC64:
+         mode64      = True;
+         getAllocableRegs_PPC ( &n_available_real_regs,
+                                &available_real_regs, mode64 );
+         isMove      = (Bool(*)(HInstr*,HReg*,HReg*)) isMove_PPCInstr;
+         getRegUsage = (void(*)(HRegUsage*,HInstr*, Bool)) getRegUsage_PPCInstr;
+         mapRegs     = (void(*)(HRegRemap*,HInstr*, Bool)) mapRegs_PPCInstr;
+         genSpill    = (HInstr*(*)(HReg,Int, Bool)) genSpill_PPC;
+         genReload   = (HInstr*(*)(HReg,Int, Bool)) genReload_PPC;
+         ppInstr     = (void(*)(HInstr*, Bool)) ppPPCInstr;
+         ppReg       = (void(*)(HReg)) ppHRegPPC;
+         iselSB      = iselSB_PPC;
+         emit        = (Int(*)(UChar*,Int,HInstr*,Bool,void*)) emit_PPCInstr;
+         host_is_bigendian = True;
+         host_word_type    = Ity_I64;
+         vassert(are_valid_hwcaps(VexArchPPC64, vta->archinfo_host.hwcaps));
+         vassert(vta->dispatch == NULL); /* return-to-dispatcher scheme */
          break;
 
       default:
@@ -331,7 +329,7 @@ VexTranslateResult LibVEX_Translate (
    }
 
 
-   switch (arch_guest) {
+   switch (vta->arch_guest) {
 
       case VexArchX86:
          preciseMemExnsFn = guest_x86_state_requires_precise_mem_exns;
@@ -342,12 +340,11 @@ VexTranslateResult LibVEX_Translate (
          guest_layout     = &x86guest_layout;
          offB_TISTART     = offsetof(VexGuestX86State,guest_TISTART);
          offB_TILEN       = offsetof(VexGuestX86State,guest_TILEN);
-         vassert(archinfo_guest->subarch == VexSubArchX86_sse0
-                 || archinfo_guest->subarch == VexSubArchX86_sse1
-                 || archinfo_guest->subarch == VexSubArchX86_sse2);
+         vassert(are_valid_hwcaps(VexArchX86, vta->archinfo_guest.hwcaps));
          vassert(0 == sizeof(VexGuestX86State) % 8);
-         vassert(sizeof( ((VexGuestX86State*)0)->guest_TISTART ) == 4);
-         vassert(sizeof( ((VexGuestX86State*)0)->guest_TILEN ) == 4);
+         vassert(sizeof( ((VexGuestX86State*)0)->guest_TISTART) == 4);
+         vassert(sizeof( ((VexGuestX86State*)0)->guest_TILEN  ) == 4);
+         vassert(sizeof( ((VexGuestX86State*)0)->guest_NRADDR ) == 4);
          break;
 
       case VexArchAMD64:
@@ -359,10 +356,11 @@ VexTranslateResult LibVEX_Translate (
          guest_layout     = &amd64guest_layout;
          offB_TISTART     = offsetof(VexGuestAMD64State,guest_TISTART);
          offB_TILEN       = offsetof(VexGuestAMD64State,guest_TILEN);
-         vassert(archinfo_guest->subarch == VexSubArch_NONE);
+         vassert(are_valid_hwcaps(VexArchAMD64, vta->archinfo_guest.hwcaps));
          vassert(0 == sizeof(VexGuestAMD64State) % 8);
          vassert(sizeof( ((VexGuestAMD64State*)0)->guest_TISTART ) == 8);
-         vassert(sizeof( ((VexGuestAMD64State*)0)->guest_TILEN ) == 8);
+         vassert(sizeof( ((VexGuestAMD64State*)0)->guest_TILEN   ) == 8);
+         vassert(sizeof( ((VexGuestAMD64State*)0)->guest_NRADDR  ) == 8);
          break;
 
       case VexArchARM:
@@ -374,24 +372,40 @@ VexTranslateResult LibVEX_Translate (
          guest_layout     = &armGuest_layout;
          offB_TISTART     = 0; /* hack ... arm has bitrot */
          offB_TILEN       = 0; /* hack ... arm has bitrot */
-         vassert(archinfo_guest->subarch == VexSubArchARM_v4);
+         vassert(are_valid_hwcaps(VexArchARM, vta->archinfo_guest.hwcaps));
          break;
 
       case VexArchPPC32:
          preciseMemExnsFn = guest_ppc32_state_requires_precise_mem_exns;
-         disInstrFn       = disInstr_PPC32;
+         disInstrFn       = disInstr_PPC;
          specHelper       = guest_ppc32_spechelper;
          guest_sizeB      = sizeof(VexGuestPPC32State);
          guest_word_type  = Ity_I32;
          guest_layout     = &ppc32Guest_layout;
          offB_TISTART     = offsetof(VexGuestPPC32State,guest_TISTART);
          offB_TILEN       = offsetof(VexGuestPPC32State,guest_TILEN);
-         vassert(archinfo_guest->subarch == VexSubArchPPC32_I
-                 || archinfo_guest->subarch == VexSubArchPPC32_FI
-                 || archinfo_guest->subarch == VexSubArchPPC32_VFI);
+         vassert(are_valid_hwcaps(VexArchPPC32, vta->archinfo_guest.hwcaps));
          vassert(0 == sizeof(VexGuestPPC32State) % 8);
          vassert(sizeof( ((VexGuestPPC32State*)0)->guest_TISTART ) == 4);
-         vassert(sizeof( ((VexGuestPPC32State*)0)->guest_TILEN ) == 4);
+         vassert(sizeof( ((VexGuestPPC32State*)0)->guest_TILEN   ) == 4);
+         vassert(sizeof( ((VexGuestPPC32State*)0)->guest_NRADDR  ) == 4);
+         break;
+
+      case VexArchPPC64:
+         preciseMemExnsFn = guest_ppc64_state_requires_precise_mem_exns;
+         disInstrFn       = disInstr_PPC;
+         specHelper       = guest_ppc64_spechelper;
+         guest_sizeB      = sizeof(VexGuestPPC64State);
+         guest_word_type  = Ity_I64;
+         guest_layout     = &ppc64Guest_layout;
+         offB_TISTART     = offsetof(VexGuestPPC64State,guest_TISTART);
+         offB_TILEN       = offsetof(VexGuestPPC64State,guest_TILEN);
+         vassert(are_valid_hwcaps(VexArchPPC64, vta->archinfo_guest.hwcaps));
+         vassert(0 == sizeof(VexGuestPPC64State) % 16);
+         vassert(sizeof( ((VexGuestPPC64State*)0)->guest_TISTART    ) == 8);
+         vassert(sizeof( ((VexGuestPPC64State*)0)->guest_TILEN      ) == 8);
+         vassert(sizeof( ((VexGuestPPC64State*)0)->guest_NRADDR     ) == 8);
+         vassert(sizeof( ((VexGuestPPC64State*)0)->guest_NRADDR_GPR2) == 8);
          break;
 
       default:
@@ -399,11 +413,11 @@ VexTranslateResult LibVEX_Translate (
    }
 
    /* yet more sanity checks ... */
-   if (arch_guest == arch_host) {
+   if (vta->arch_guest == vta->arch_host) {
       /* doesn't necessarily have to be true, but if it isn't it means
          we are simulating one flavour of an architecture a different
          flavour of the same architecture, which is pretty strange. */
-      vassert(archinfo_guest->subarch == archinfo_host->subarch);
+      vassert(vta->archinfo_guest.hwcaps == vta->archinfo_host.hwcaps);
    }
 
    vexAllocSanityCheck();
@@ -413,100 +427,106 @@ VexTranslateResult LibVEX_Translate (
                    " Front end "
                    "------------------------\n\n");
 
-   irbb = bb_to_IR ( guest_extents,
+   irsb = bb_to_IR ( vta->guest_extents,
+                     vta->callback_opaque,
                      disInstrFn,
-                     guest_bytes, 
-                     guest_bytes_addr,
-                     chase_into_ok,
+                     vta->guest_bytes, 
+                     vta->guest_bytes_addr,
+                     vta->chase_into_ok,
                      host_is_bigendian,
-                     archinfo_guest,
+                     vta->arch_guest,
+                     &vta->archinfo_guest,
+                     &vta->abiinfo_both,
                      guest_word_type,
-                     do_self_check,
+                     vta->do_self_check,
+                     vta->preamble_function,
                      offB_TISTART,
                      offB_TILEN );
 
    vexAllocSanityCheck();
 
-   if (irbb == NULL) {
+   if (irsb == NULL) {
       /* Access failure. */
       vexSetAllocModeTEMP_and_clear();
       vex_traceflags = 0;
       return VexTransAccessFail;
    }
 
-   vassert(guest_extents->n_used >= 1 && guest_extents->n_used <= 3);
-   vassert(guest_extents->base[0] == guest_bytes_addr);
-   for (i = 0; i < guest_extents->n_used; i++) {
-      vassert(guest_extents->len[i] < 10000); /* sanity */
+   vassert(vta->guest_extents->n_used >= 1 && vta->guest_extents->n_used <= 3);
+   vassert(vta->guest_extents->base[0] == vta->guest_bytes_addr);
+   for (i = 0; i < vta->guest_extents->n_used; i++) {
+      vassert(vta->guest_extents->len[i] < 10000); /* sanity */
    }
 
    /* If debugging, show the raw guest bytes for this bb. */
    if (0 || (vex_traceflags & VEX_TRACE_FE)) {
-      if (guest_extents->n_used > 1) {
+      if (vta->guest_extents->n_used > 1) {
          vex_printf("can't show code due to extents > 1\n");
       } else {
          /* HACK */
-         UChar* p = (UChar*)guest_bytes;
-         UInt   guest_bytes_read = (UInt)guest_extents->len[0];
-         vex_printf(". 0 %llx %u\n.", guest_bytes_addr, guest_bytes_read );
+         UChar* p = (UChar*)vta->guest_bytes;
+         UInt   guest_bytes_read = (UInt)vta->guest_extents->len[0];
+         vex_printf(". 0 %llx %u\n.", vta->guest_bytes_addr, guest_bytes_read );
          for (i = 0; i < guest_bytes_read; i++)
-         vex_printf(" %02x", (Int)p[i] );
+            vex_printf(" %02x", (Int)p[i] );
          vex_printf("\n\n");
       }
    }
 
    /* Sanity check the initial IR. */
-   sanityCheckIRBB( irbb, "initial IR", 
+   sanityCheckIRSB( irsb, "initial IR", 
                     False/*can be non-flat*/, guest_word_type );
 
    vexAllocSanityCheck();
 
    /* Clean it up, hopefully a lot. */
-   irbb = do_iropt_BB ( irbb, specHelper, preciseMemExnsFn, 
-                              guest_bytes_addr );
-   sanityCheckIRBB( irbb, "after initial iropt", 
+   irsb = do_iropt_BB ( irsb, specHelper, preciseMemExnsFn, 
+                              vta->guest_bytes_addr );
+   sanityCheckIRSB( irsb, "after initial iropt", 
                     True/*must be flat*/, guest_word_type );
 
    if (vex_traceflags & VEX_TRACE_OPT1) {
       vex_printf("\n------------------------" 
                    " After pre-instr IR optimisation "
                    "------------------------\n\n");
-      ppIRBB ( irbb );
+      ppIRSB ( irsb );
       vex_printf("\n");
    }
 
    vexAllocSanityCheck();
 
    /* Get the thing instrumented. */
-   if (instrument1)
-      irbb = (*instrument1)(irbb, guest_layout, 
-                            guest_bytes_addr_noredir, guest_extents,
-                            guest_word_type, host_word_type);
+   if (vta->instrument1)
+      irsb = vta->instrument1(vta->callback_opaque,
+                              irsb, guest_layout, 
+                              vta->guest_extents,
+                              guest_word_type, host_word_type);
    vexAllocSanityCheck();
 
-   if (instrument2)
-      irbb = (*instrument2)(irbb, guest_layout,
-                            guest_bytes_addr_noredir, guest_extents,
-                            guest_word_type, host_word_type);
+   if (vta->instrument2)
+      irsb = vta->instrument2(vta->callback_opaque,
+                              irsb, guest_layout,
+                              vta->guest_extents,
+                              guest_word_type, host_word_type);
       
    if (vex_traceflags & VEX_TRACE_INST) {
       vex_printf("\n------------------------" 
                    " After instrumentation "
                    "------------------------\n\n");
-      ppIRBB ( irbb );
+      ppIRSB ( irsb );
       vex_printf("\n");
    }
 
-   if (instrument1 || instrument2)
-      sanityCheckIRBB( irbb, "after instrumentation",
+   if (vta->instrument1 || vta->instrument2)
+      sanityCheckIRSB( irsb, "after instrumentation",
                        True/*must be flat*/, guest_word_type );
 
    /* Do a post-instrumentation cleanup pass. */
-   if (cleanup_after_instrumentation) {
-      do_deadcode_BB( irbb );
-      irbb = cprop_BB( irbb );
-      do_deadcode_BB( irbb );
-      sanityCheckIRBB( irbb, "after post-instrumentation cleanup",
+   if (vta->instrument1 || vta->instrument2) {
+      do_deadcode_BB( irsb );
+      irsb = cprop_BB( irsb );
+      do_deadcode_BB( irsb );
+      sanityCheckIRSB( irsb, "after post-instrumentation cleanup",
                        True/*must be flat*/, guest_word_type );
    }
 
@@ -516,13 +536,17 @@ VexTranslateResult LibVEX_Translate (
       vex_printf("\n------------------------" 
                    " After post-instr IR optimisation "
                    "------------------------\n\n");
-      ppIRBB ( irbb );
+      ppIRSB ( irsb );
       vex_printf("\n");
    }
 
-   /* Turn it into virtual-registerised code. */
-   do_deadcode_BB( irbb );
-   do_treebuild_BB( irbb );
+   /* Turn it into virtual-registerised code.  Build trees -- this
+      also throws away any dead bindings. */
+   ado_treebuild_BB( irsb );
+
+   if (vta->finaltidy) {
+      irsb = vta->finaltidy(irsb);
+   }
 
    vexAllocSanityCheck();
 
@@ -530,12 +554,12 @@ VexTranslateResult LibVEX_Translate (
       vex_printf("\n------------------------" 
                    "  After tree-building "
                    "------------------------\n\n");
-      ppIRBB ( irbb );
+      ppIRSB ( irsb );
       vex_printf("\n");
    }
 
    /* HACK */
-   if (0) { *host_bytes_used = 0; return VexTransOK; }
+   if (0) { *(vta->host_bytes_used) = 0; return VexTransOK; }
    /* end HACK */
 
    if (vex_traceflags & VEX_TRACE_VCODE)
@@ -543,7 +567,8 @@ VexTranslateResult LibVEX_Translate (
                    " Instruction selection "
                    "------------------------\n");
 
-   vcode = iselBB ( irbb, archinfo_host );
+   vcode = iselSB ( irsb, vta->arch_host, &vta->archinfo_host, 
+                                          &vta->abiinfo_both );
 
    vexAllocSanityCheck();
 
@@ -553,7 +578,7 @@ VexTranslateResult LibVEX_Translate (
    if (vex_traceflags & VEX_TRACE_VCODE) {
       for (i = 0; i < vcode->arr_used; i++) {
          vex_printf("%3d   ", i);
-         ppInstr(vcode->arr[i]);
+         ppInstr(vcode->arr[i], mode64);
          vex_printf("\n");
       }
       vex_printf("\n");
@@ -561,10 +586,11 @@ VexTranslateResult LibVEX_Translate (
 
    /* Register allocate. */
    rcode = doRegisterAllocation ( vcode, available_real_regs,
-                                           n_available_real_regs,
+                                  n_available_real_regs,
                                   isMove, getRegUsage, mapRegs, 
-                                  genSpill, genReload, guest_sizeB,
-                                  ppInstr, ppReg );
+                                  genSpill, genReload, directReload, 
+                                  guest_sizeB,
+                                  ppInstr, ppReg, mode64 );
 
    vexAllocSanityCheck();
 
@@ -574,14 +600,14 @@ VexTranslateResult LibVEX_Translate (
                    "------------------------\n\n");
       for (i = 0; i < rcode->arr_used; i++) {
          vex_printf("%3d   ", i);
-         ppInstr(rcode->arr[i]);
+         ppInstr(rcode->arr[i], mode64);
          vex_printf("\n");
       }
       vex_printf("\n");
    }
 
    /* HACK */
-   if (0) { *host_bytes_used = 0; return VexTransOK; }
+   if (0) { *(vta->host_bytes_used) = 0; return VexTransOK; }
    /* end HACK */
 
    /* Assemble */
@@ -594,10 +620,10 @@ VexTranslateResult LibVEX_Translate (
    out_used = 0; /* tracks along the host_bytes array */
    for (i = 0; i < rcode->arr_used; i++) {
       if (vex_traceflags & VEX_TRACE_ASM) {
-         ppInstr(rcode->arr[i]);
+         ppInstr(rcode->arr[i], mode64);
          vex_printf("\n");
       }
-      j = (*emit)( insn_bytes, 32, rcode->arr[i] );
+      j = (*emit)( insn_bytes, 32, rcode->arr[i], mode64, vta->dispatch );
       if (vex_traceflags & VEX_TRACE_ASM) {
          for (k = 0; k < j; k++)
             if (insn_bytes[k] < 16)
@@ -606,18 +632,18 @@ VexTranslateResult LibVEX_Translate (
                vex_printf("%x ", (UInt)insn_bytes[k]);
          vex_printf("\n\n");
       }
-      if (out_used + j > host_bytes_size) {
+      if (out_used + j > vta->host_bytes_size) {
          vexSetAllocModeTEMP_and_clear();
          vex_traceflags = 0;
          return VexTransOutputFull;
       }
       for (k = 0; k < j; k++) {
-         host_bytes[out_used] = insn_bytes[k];
+         vta->host_bytes[out_used] = insn_bytes[k];
          out_used++;
       }
-      vassert(out_used <= host_bytes_size);
+      vassert(out_used <= vta->host_bytes_size);
    }
-   *host_bytes_used = out_used;
+   *(vta->host_bytes_used) = out_used;
 
    vexAllocSanityCheck();
 
@@ -647,14 +673,18 @@ HChar* LibVEX_EmWarn_string ( VexEmWarn ew )
         return "Setting %mxcsr.daz (SSE treat-denormals-as-zero mode)";
      case EmWarn_X86_acFlag:
         return "Setting %eflags.ac (setting noted but ignored)";
-     case EmWarn_PPC32exns:
-        return "Unmasking PPC32 FP exceptions";
+     case EmWarn_PPCexns:
+        return "Unmasking PPC32/64 FP exceptions";
+     case EmWarn_PPC64_redir_overflow:
+        return "PPC64 function redirection stack overflow";
+     case EmWarn_PPC64_redir_underflow:
+        return "PPC64 function redirection stack underflow";
      default: 
         vpanic("LibVEX_EmWarn_string: unknown warning");
    }
 }
 
-/* --------- Arch/Subarch stuff. --------- */
+/* ------------------ Arch/HwCaps stuff. ------------------ */
 
 const HChar* LibVEX_ppVexArch ( VexArch arch )
 {
@@ -664,31 +694,129 @@ const HChar* LibVEX_ppVexArch ( VexArch arch )
       case VexArchAMD64:    return "AMD64";
       case VexArchARM:      return "ARM";
       case VexArchPPC32:    return "PPC32";
+      case VexArchPPC64:    return "PPC64";
       default:              return "VexArch???";
    }
 }
 
-const HChar* LibVEX_ppVexSubArch ( VexSubArch subarch )
+const HChar* LibVEX_ppVexHwCaps ( VexArch arch, UInt hwcaps )
 {
-   switch (subarch) {
-      case VexSubArch_INVALID:   return "INVALID";
-      case VexSubArch_NONE:      return "NONE";
-      case VexSubArchX86_sse0:   return "x86-sse0";
-      case VexSubArchX86_sse1:   return "x86-sse1";
-      case VexSubArchX86_sse2:   return "x86-sse2";
-      case VexSubArchARM_v4:     return "arm-v4";
-      case VexSubArchPPC32_I:    return "ppc32-int-only";
-      case VexSubArchPPC32_FI:   return "ppc32-int-and-fp";
-      case VexSubArchPPC32_VFI:  return "ppc32-int-fp-and-AV";
-      default:                   return "VexSubArch???";
-   }
+   HChar* str = show_hwcaps(arch,hwcaps);
+   return str ? str : "INVALID";
 }
 
+
 /* Write default settings info *vai. */
 void LibVEX_default_VexArchInfo ( /*OUT*/VexArchInfo* vai )
 {
-   vai->subarch              = VexSubArch_INVALID;
-   vai->ppc32_cache_line_szB = 0;
+   vai->hwcaps             = 0;
+   vai->ppc_cache_line_szB = 0;
+}
+
+/* Write default settings info *vbi. */
+void LibVEX_default_VexAbiInfo ( /*OUT*/VexAbiInfo* vbi )
+{
+   vbi->guest_stack_redzone_size       = 0;
+   vbi->guest_ppc_zap_RZ_at_blr        = False;
+   vbi->guest_ppc_zap_RZ_at_bl         = NULL;
+   vbi->guest_ppc_sc_continues_at_LR   = False;
+   vbi->host_ppc_calls_use_fndescrs    = False;
+   vbi->host_ppc32_regalign_int64_args = False;
+}
+
+
+/* Return a string showing the hwcaps in a nice way.  The string will
+   be NULL for invalid combinations of flags, so these functions also
+   serve as a way to validate hwcaps values. */
+
+static HChar* show_hwcaps_x86 ( UInt hwcaps ) 
+{
+   /* Monotonic, SSE3 > SSE2 > SSE1 > baseline. */
+   if (hwcaps == 0)
+      return "x86-sse0";
+   if (hwcaps == VEX_HWCAPS_X86_SSE1)
+      return "x86-sse1";
+   if (hwcaps == (VEX_HWCAPS_X86_SSE1 | VEX_HWCAPS_X86_SSE2))
+      return "x86-sse1-sse2";
+   if (hwcaps == (VEX_HWCAPS_X86_SSE1 
+                  | VEX_HWCAPS_X86_SSE2 | VEX_HWCAPS_X86_SSE3))
+      return "x86-sse1-sse2-sse3";
+
+   return False;
+}
+
+static HChar* show_hwcaps_amd64 ( UInt hwcaps )
+{
+   /* Monotonic, SSE3 > baseline. */
+   if (hwcaps == 0)
+      return "amd64-sse2";
+   if (hwcaps == VEX_HWCAPS_AMD64_SSE3)
+      return "amd64-sse3";
+   return False;
+}
+
+static HChar* show_hwcaps_ppc32 ( UInt hwcaps )
+{
+   /* Monotonic with complications.  Basically V > F > baseline,
+      but once you have F then you can have FX or GX too. */
+   const UInt F  = VEX_HWCAPS_PPC32_F;
+   const UInt V  = VEX_HWCAPS_PPC32_V;
+   const UInt FX = VEX_HWCAPS_PPC32_FX;
+   const UInt GX = VEX_HWCAPS_PPC32_GX;
+         UInt c  = hwcaps;
+   if (c == 0)           return "ppc32-int";
+   if (c == F)           return "ppc32-int-flt";
+   if (c == (F|FX))      return "ppc32-int-flt-FX";
+   if (c == (F|GX))      return "ppc32-int-flt-GX";
+   if (c == (F|FX|GX))   return "ppc32-int-flt-FX-GX";
+   if (c == (F|V))       return "ppc32-int-flt-vmx";
+   if (c == (F|V|FX))    return "ppc32-int-flt-vmx-FX";
+   if (c == (F|V|GX))    return "ppc32-int-flt-vmx-GX";
+   if (c == (F|V|FX|GX)) return "ppc32-int-flt-vmx-FX-GX";
+   return NULL;
+}
+
+static HChar* show_hwcaps_ppc64 ( UInt hwcaps )
+{
+   /* Monotonic with complications.  Basically V > baseline(==F),
+      but once you have F then you can have FX or GX too. */
+   const UInt V  = VEX_HWCAPS_PPC64_V;
+   const UInt FX = VEX_HWCAPS_PPC64_FX;
+   const UInt GX = VEX_HWCAPS_PPC64_GX;
+         UInt c  = hwcaps;
+   if (c == 0)         return "ppc64-int-flt";
+   if (c == FX)        return "ppc64-int-flt-FX";
+   if (c == GX)        return "ppc64-int-flt-GX";
+   if (c == (FX|GX))   return "ppc64-int-flt-FX-GX";
+   if (c == V)         return "ppc64-int-flt-vmx";
+   if (c == (V|FX))    return "ppc64-int-flt-vmx-FX";
+   if (c == (V|GX))    return "ppc64-int-flt-vmx-GX";
+   if (c == (V|FX|GX)) return "ppc64-int-flt-vmx-FX-GX";
+   return NULL;
+}
+
+static HChar* show_hwcaps_arm ( UInt hwcaps )
+{
+   if (hwcaps == 0) return "arm-baseline";
+   return NULL;
+}
+
+/* ---- */
+static HChar* show_hwcaps ( VexArch arch, UInt hwcaps )
+{
+   switch (arch) {
+      case VexArchX86:   return show_hwcaps_x86(hwcaps);
+      case VexArchAMD64: return show_hwcaps_amd64(hwcaps);
+      case VexArchPPC32: return show_hwcaps_ppc32(hwcaps);
+      case VexArchPPC64: return show_hwcaps_ppc64(hwcaps);
+      case VexArchARM:   return show_hwcaps_arm(hwcaps);
+      default: return NULL;
+   }
+}
+
+static Bool are_valid_hwcaps ( VexArch arch, UInt hwcaps )
+{
+   return show_hwcaps(arch,hwcaps) != NULL;
 }
 
 
