@@ -86,11 +86,21 @@
    Open MPI      lib/libmpi.so,   soname = libmpi.so.0
    Quadrics MPI  lib/libmpi.so,   soname = libmpi.so.0
    MPICH         libmpich.so.1.0, soname = libmpich.so.1.0
+   AIX: in /usr/lpp/ppe.poe/lib/libmpi_r.a(mpicore*_r.o)
 
-   So a suitable soname to match with is "libmpi*.so*".
+   For the non-AIX targets, a suitable soname to match with
+   is "libmpi*.so*".
 */
-/* ifdef OpenMPI ... */
-#define I_WRAP_FNNAME_U(_name) I_WRAP_SONAME_FNNAME_ZU(libmpiZaZdsoZa,_name)
+#if defined(_AIX)
+# define I_WRAP_FNNAME_U(_name) \
+         I_WRAP_SONAME_FNNAME_ZU(libmpiZurZdaZLmpicoreZaZurZdoZR,_name)
+  /* Don't change this without also changing all the names in
+     libmpiwrap.exp. */
+#else
+# define I_WRAP_FNNAME_U(_name) \
+         I_WRAP_SONAME_FNNAME_ZU(libmpiZaZdsoZa,_name)
+
+#endif
 
 
 /*------------------------------------------------------------*/
@@ -126,11 +136,12 @@ static long sizeof_long_double_image ( void );
 static const char* preamble = "valgrind MPI wrappers";
 
 /* established at startup */
-static pid_t my_pid        = -1;
-static char* options_str   = NULL;
-static int   opt_verbosity = 1;
-static Bool  opt_missing   = 0; /* 0:silent; 1:warn; 2:abort */
-static Bool  opt_help      = False;
+static pid_t my_pid         = -1;
+static char* options_str    = NULL;
+static int   opt_verbosity  = 1;
+static Bool  opt_missing    = 0; /* 0:silent; 1:warn; 2:abort */
+static Bool  opt_help       = False;
+static Bool  opt_initkludge = False;
 
 static void before ( char* fnname )
 {
@@ -152,6 +163,8 @@ static void before ( char* fnname )
             opt_verbosity--;
          if (NULL != strstr(options_str, "help"))
             opt_help = True;
+         if (NULL != strstr(options_str, "initkludge"))
+            opt_initkludge = True;
       }
       if (opt_verbosity > 0)
          fprintf(stderr, "%s %5d: Active for pid %d\n", 
@@ -167,13 +180,14 @@ static void before ( char* fnname )
          fprintf(stderr, "Valid options for the MPIWRAP_DEBUG environment"
                          " variable are:\n");
          fprintf(stderr, "\n");
-         fprintf(stderr, "   quiet      be silent except for errors\n");
-         fprintf(stderr, "   verbose    show wrapper entries/exits\n");
-         fprintf(stderr, "   strict     abort the program if a function"
+         fprintf(stderr, "   quiet       be silent except for errors\n");
+         fprintf(stderr, "   verbose     show wrapper entries/exits\n");
+         fprintf(stderr, "   strict      abort the program if a function"
                          " with no wrapper is used\n");
-         fprintf(stderr, "   warn       give a warning if a function"
+         fprintf(stderr, "   warn        give a warning if a function"
                          " with no wrapper is used\n");
-         fprintf(stderr, "   help       display this message, then exit\n");
+         fprintf(stderr, "   help        display this message, then exit\n");
+         fprintf(stderr, "   initkludge  debugging hack; do not use\n");
          fprintf(stderr, "\n");
          fprintf(stderr, "Multiple options are allowed, eg"
                          " MPIWRAP_DEBUG=strict,verbose\n");
@@ -193,7 +207,7 @@ static void before ( char* fnname )
       fprintf(stderr, "%s %5d: enter PMPI_%s\n", preamble,  my_pid, fnname );
 }
 
-static inline void after ( char* fnname, int err )
+static __inline__ void after ( char* fnname, int err )
 {
    if (opt_verbosity > 1)
       fprintf(stderr, "%s %5d:  exit PMPI_%s (err = %d)\n", 
@@ -241,6 +255,54 @@ static void showTy ( FILE* f, MPI_Datatype ty )
 #  endif
 #  if defined(MPI_UNSIGNED_LONG_LONG)
    else if (ty == MPI_UNSIGNED_LONG_LONG) fprintf(f,"UNSIGNED_LONG_LONG");
+#  endif
+#  if defined(MPI_REAL8)
+   else if (ty == MPI_REAL8)          fprintf(f, "REAL8");
+#  endif
+#  if defined(MPI_REAL4)
+   else if (ty == MPI_REAL4)          fprintf(f, "REAL4");
+#  endif
+#  if defined(MPI_REAL)
+   else if (ty == MPI_REAL)           fprintf(f, "REAL");
+#  endif
+#  if defined(MPI_INTEGER8)
+   else if (ty == MPI_INTEGER8)       fprintf(f, "INTEGER8");
+#  endif
+#  if defined(MPI_INTEGER4)
+   else if (ty == MPI_INTEGER4)       fprintf(f, "INTEGER4");
+#  endif
+#  if defined(MPI_INTEGER)
+   else if (ty == MPI_INTEGER)        fprintf(f, "INTEGER");
+#  endif
+#  if defined(MPI_DOUBLE_PRECISION)
+   else if (ty == MPI_DOUBLE_PRECISION) fprintf(f, "DOUBLE_PRECISION");
+#  endif
+#  if defined(MPI_COMPLEX)
+   else if (ty == MPI_COMPLEX)          fprintf(f, "COMPLEX");
+#  endif
+#  if defined(MPI_DOUBLE_COMPLEX)
+   else if (ty == MPI_DOUBLE_COMPLEX)   fprintf(f, "DOUBLE_COMPLEX");
+#  endif
+#  if defined(MPI_LOGICAL)
+   else if (ty == MPI_LOGICAL)          fprintf(f, "LOGICAL");
+#  endif
+#  if defined(MPI_2INTEGER)
+   else if (ty == MPI_2INTEGER)         fprintf(f, "2INTEGER");
+#  endif
+#  if defined(MPI_2COMPLEX)
+   else if (ty == MPI_2COMPLEX)         fprintf(f, "2COMPLEX");
+#  endif
+#  if defined(MPI_2DOUBLE_COMPLEX)
+   else if (ty == MPI_2DOUBLE_COMPLEX)  fprintf(f, "2DOUBLE_COMPLEX");
+#  endif
+#  if defined(MPI_2REAL)
+   else if (ty == MPI_2REAL)            fprintf(f, "2REAL");
+#  endif
+#  if defined(MPI_2DOUBLE_PRECISION)
+   else if (ty == MPI_2DOUBLE_PRECISION) fprintf(f, "2DOUBLE_PRECISION");
+#  endif
+#  if defined(MPI_CHARACTER)
+   else if (ty == MPI_CHARACTER)         fprintf(f, "CHARACTER");
 #  endif
    else fprintf(f,"showTy:???");
 }
@@ -298,21 +360,21 @@ static void showCombiner ( FILE* f, int combiner )
 /* Note, PMPI_Comm_rank/size are themselves wrapped.  Should work
    fine. */
 
-static inline int comm_rank ( MPI_Comm comm ) 
+static __inline__ int comm_rank ( MPI_Comm comm ) 
 {
    int err, r;
    err = PMPI_Comm_rank(comm, &r);
    return err ? 0/*arbitrary*/ : r;
 }
 
-static inline int comm_size ( MPI_Comm comm ) 
+static __inline__ int comm_size ( MPI_Comm comm ) 
 {
    int err, r;
    err = PMPI_Comm_size(comm, &r);
    return err ? 0/*arbitrary*/ : r;
 }
 
-static inline Bool count_from_Status( /*OUT*/int* recv_count, 
+static __inline__ Bool count_from_Status( /*OUT*/int* recv_count, 
                                       MPI_Datatype datatype, 
                                       MPI_Status* status)
 {
@@ -333,7 +395,7 @@ static inline Bool count_from_Status( /*OUT*/int* recv_count,
    types that support assignment and equality operations."  Hence the
    following function should compile for any compliant definition of
    MPI_Request. */
-static inline 
+static __inline__ 
 Bool eq_MPI_Request ( MPI_Request r1, MPI_Request r2 )
 {
    return r1 == r2;
@@ -410,19 +472,71 @@ static long sizeofOneNamedTy ( MPI_Datatype ty )
    if (ty == MPI_DOUBLE)         return sizeof(double);
    if (ty == MPI_BYTE)           return 1;
    if (ty == MPI_LONG_DOUBLE)    return sizeof_long_double_image();
+   if (ty == MPI_PACKED)         return 1;
+   if (ty == MPI_LONG_LONG_INT)  return sizeof(signed long long int);
+
+#  if defined(MPI_REAL8)
+   if (ty == MPI_REAL8)          return 8; /* MPI2 spec */;
+#  endif
+#  if defined(MPI_REAL4)
+   if (ty == MPI_REAL4)          return 4; /* MPI2 spec */;
+#  endif
+#  if defined(MPI_REAL)
+   if (ty == MPI_REAL)           return 4; /* MPI2 spec */;
+#  endif
+#  if defined(MPI_INTEGER8)
+   if (ty == MPI_INTEGER8)       return 8; /* MPI2 spec */;
+#  endif
+#  if defined(MPI_INTEGER4)
+   if (ty == MPI_INTEGER4)       return 4; /* MPI2 spec */;
+#  endif
+#  if defined(MPI_INTEGER)
+   if (ty == MPI_INTEGER)        return 4; /* MPI2 spec */;
+#  endif
+#  if defined(MPI_DOUBLE_PRECISION)
+   if (ty == MPI_DOUBLE_PRECISION) return 8; /* MPI2 spec */;
+#  endif
 
-   /* MPI_PACKED */
    /* new in MPI2: */
 #  if defined(MPI_WCHAR)
-   if (ty == MPI_WCHAR)              return sizeof(wchar_t);
+   if (ty == MPI_WCHAR)              return 2; /* MPI2 spec */;
 #  endif
 #  if defined(MPI_SIGNED_CHAR)
-   if (ty == MPI_SIGNED_CHAR)        return sizeof(signed char);
+   if (ty == MPI_SIGNED_CHAR)        return 1; /* MPI2 spec */;
 #  endif
 #  if defined(MPI_UNSIGNED_LONG_LONG)
-   if (ty == MPI_UNSIGNED_LONG_LONG) return sizeof(unsigned long long int);
+   if (ty == MPI_UNSIGNED_LONG_LONG) return 8; /* MPI2 spec */;
 #  endif
-   if (ty == MPI_LONG_LONG_INT)      return sizeof(signed long long int);
+#  if defined(MPI_COMPLEX)
+   if (ty == MPI_COMPLEX)            return 2 * 4; /* MPI2 spec */
+#  endif
+#  if defined(MPI_DOUBLE_COMPLEX)
+   if (ty == MPI_DOUBLE_COMPLEX)     return 2 * 8; /* MPI2 spec */
+#  endif
+#  if defined(MPI_LOGICAL)
+   if (ty == MPI_LOGICAL)            return 4; /* MPI2 spec */
+#  endif
+#  if defined(MPI_2INTEGER)
+   if (ty == MPI_2INTEGER)      return 2 * 4; /* undocumented in MPI2 */
+#  endif
+#  if defined(MPI_2COMPLEX)
+   if (ty == MPI_2COMPLEX)      return 2 * 8; /* undocumented in MPI2 */
+#  endif
+#  if defined(MPI_2DOUBLE_COMPLEX)
+   /* 32: this is how openmpi-1.2.2 behaves on x86-linux, but I have
+      really no idea if this is right. */
+   if (ty == MPI_2DOUBLE_COMPLEX)   return 32; /* undocumented in MPI2 */
+#  endif
+#  if defined(MPI_2REAL)
+   if (ty == MPI_2REAL)              return 2 * 4; /* undocumented in MPI2 */
+#  endif
+#  if defined(MPI_2DOUBLE_PRECISION)
+   if (ty == MPI_2DOUBLE_PRECISION)  return 2 * 8; /* undocumented in MPI2 */
+#  endif
+#  if defined(MPI_CHARACTER)
+   if (ty == MPI_CHARACTER)          return 1; /* MPI2 spec */
+#  endif
+
    /* Note: the following are named structs, not named basic types,
       and so are not handled here:
          FLOAT_INT DOUBLE_INT LONG_INT 2INT SHORT_INT LONG_DOUBLE_INT
@@ -509,6 +623,10 @@ void walk_type ( void(*f)(void*,long), char* base, MPI_Datatype ty )
    MPI_Aint*     addrs = NULL;
    MPI_Datatype* dtys  = NULL;
 
+   /* Stuff for limiting how much complaining text it spews out */
+   static int complaints = 3;
+   static int last_complained_about_tycon = -987654321; /* presumably bogus */
+
    if (0)
       printf("walk_type %p\n", (void*)(unsigned long)ty);
 
@@ -526,8 +644,14 @@ void walk_type ( void(*f)(void*,long), char* base, MPI_Datatype ty )
          take them to bits so we have to do a really ugly hack, which
          makes assumptions about how the MPI implementation has laid
          out these types.  At least Open MPI 1.0.1 appears to put
-         the 'val' field first.
+         the 'val' field first.  MPICH2 agrees.
       */
+      if (ty == MPI_2INT) {
+         typedef struct { int val; int loc; } Ty;
+         f(base + offsetof(Ty,val), sizeof(int));
+         f(base + offsetof(Ty,loc), sizeof(int));
+         return;
+      }
       if (ty == MPI_LONG_INT) {
          typedef struct { long val; int loc; } Ty;
          f(base + offsetof(Ty,val), sizeof(long));
@@ -546,6 +670,18 @@ void walk_type ( void(*f)(void*,long), char* base, MPI_Datatype ty )
          f(base + offsetof(Ty,loc), sizeof(int));
          return;
       }
+      if (ty == MPI_FLOAT_INT) {
+         typedef struct { float val; int loc; } Ty;
+         f(base + offsetof(Ty,val), sizeof(float));
+         f(base + offsetof(Ty,loc), sizeof(int));
+         return;
+      }
+      if (ty == MPI_LONG_DOUBLE_INT) {
+         typedef struct { long double val; int loc; } Ty;
+         f(base + offsetof(Ty,val), sizeof_long_double_image());
+         f(base + offsetof(Ty,loc), sizeof(int));
+         return;
+      }
       if (ty == MPI_LB || ty == MPI_UB)
          return; /* have zero size, so nothing needs to be done */
       goto unhandled;
@@ -671,14 +807,21 @@ void walk_type ( void(*f)(void*,long), char* base, MPI_Datatype ty )
    return;
 
   unhandled:
-   if (tycon == MPI_COMBINER_NAMED) {
-      fprintf(stderr, "%s %5d: walk_type: unhandled base type 0x%lx ",
-                      preamble, my_pid, (long)ty);
-      showTy(stderr, ty);
-      fprintf(stderr, "\n");
-   } else {
-      fprintf(stderr, "%s %5d: walk_type: unhandled combiner 0x%lx\n",
-                      preamble, my_pid, (long)tycon);
+   /* Complain, but limit the amount of complaining that can happen to
+      the first 3 different unhandled tycons that show up, so as to
+      avoid swamping users with thousands of duplicate messages. */
+   if (complaints > 0 && tycon != last_complained_about_tycon) {
+      complaints--;
+      last_complained_about_tycon = tycon;
+      if (tycon == MPI_COMBINER_NAMED) {
+         fprintf(stderr, "%s %5d: walk_type: unhandled base type 0x%lx ",
+                         preamble, my_pid, (long)ty);
+         showTy(stderr, ty);
+         fprintf(stderr, "\n");
+      } else {
+         fprintf(stderr, "%s %5d: walk_type: unhandled combiner 0x%lx\n",
+                         preamble, my_pid, (long)tycon);
+      }
    }
    if (ints)  free(ints);
    if (addrs) free(addrs);
@@ -738,7 +881,7 @@ void walk_type_array ( void(*f)(void*,long), char* base,
 void mpiwrap_walk_type_EXTERNALLY_VISIBLE
     ( void(*f)(void*,long), char* base, MPI_Datatype ty )
 {
-   return walk_type(f, base, ty);
+   walk_type(f, base, ty);
 }
 
 
@@ -927,7 +1070,11 @@ int WRAPPER_FOR(PMPI_Get_count)(MPI_Status* status,
    int    err;
    VALGRIND_GET_ORIG_FN(fn);
    before("Get_count");
+#  if defined(_AIX)
+   check_mem_is_addressable_untyped(status, sizeof(*status));
+#  else
    check_mem_is_defined_untyped(status, sizeof(*status));
+#  endif
    CALL_FN_W_WWW(err, fn, status,ty,count);
    after("Get_count", err);
    return err;
@@ -1114,7 +1261,7 @@ MPI_Request* clone_Request_array ( int count, MPI_Request* orig )
    if (count < 0) 
       count = 0; /* Hmm.  Call Mulder and Scully. */
    copy = malloc( count * sizeof(MPI_Request) );
-   if (copy == NULL) {
+   if (copy == NULL && count > 0) {
       UNLOCK_SREQS;
       barf("clone_Request_array: malloc failed");
    }
@@ -1493,6 +1640,107 @@ int WRAPPER_FOR(PMPI_Type_free)( MPI_Datatype* ty )
 }
 
 
+/*------------------------------------------------------------*/
+/*---                                                      ---*/
+/*--- Sec 3.13, Pack and unpack                            ---*/
+/*---                                                      ---*/
+/*------------------------------------------------------------*/
+
+/* --- Pack --- */
+/* pre: must be readable: position
+        must be readable: (inbuf,incount,datatype)
+        must be writable: outbuf[0 .. outsize-1]
+        must be writable: outbuf[*position .. 
+                                 *position - 1 
+                                 + however much space PMPI_Pack_size 
+                                   says we will need]
+   post: make readable: outbuf[old *position .. new *position]
+*/
+int WRAPPER_FOR(PMPI_Pack)( void* inbuf, int incount, MPI_Datatype datatype, 
+                            void* outbuf, int outsize, 
+                            int* position, MPI_Comm comm ) 
+{
+   OrigFn fn;
+   int    err, szB = 0;
+   int    position_ORIG = *position;
+   VALGRIND_GET_ORIG_FN(fn);
+   before("Pack");
+   /* stay sane */
+   check_mem_is_defined_untyped(position, sizeof(*position));
+   /* check input */
+   check_mem_is_defined(inbuf, incount, datatype);
+   /* check output area's stated bounds make sense */
+   check_mem_is_addressable_untyped(outbuf, outsize);
+   /* check output area's actual used size properly */
+   err = PMPI_Pack_size( incount, datatype, comm, &szB );
+   if (err == MPI_SUCCESS && szB > 0) {
+      check_mem_is_addressable_untyped( 
+         ((char*)outbuf) + position_ORIG, szB
+      );
+   }
+
+   CALL_FN_W_7W(err, fn, inbuf,incount,datatype, outbuf,outsize,position, comm);
+
+   if (err == MPI_SUCCESS && (*position) > position_ORIG) {
+      /* paint output */
+      make_mem_defined_if_addressable_untyped( 
+         ((char*)outbuf) + position_ORIG, *position - position_ORIG
+      );
+   }
+   after("Pack", err);
+   return err;
+}
+
+/* --- Unpack --- */
+/* pre: must be readable: position
+        must be writable: (outbuf,outcount,datatype)
+        must be writable: outbuf[0 .. outsize-1]
+        must be writable: outbuf[*position .. 
+                                 *position - 1 
+                                 + however much space PMPI_Pack_size 
+                                   says we will need]
+   post: make readable: (outbuf,outcount,datatype)
+         and also do a readability check of
+         inbuf[old *position .. new *position]
+*/
+int WRAPPER_FOR(PMPI_Unpack)( void* inbuf, int insize, int* position,
+                              void* outbuf, int outcount, MPI_Datatype datatype, 
+                              MPI_Comm comm )
+{
+   OrigFn fn;
+   int    err, szB = 0;
+   int    position_ORIG = *position;
+   VALGRIND_GET_ORIG_FN(fn);
+   before("Unpack");
+   /* stay sane */
+   check_mem_is_defined_untyped(position, sizeof(*position));
+   /* check output area is accessible */
+   check_mem_is_addressable(outbuf, outcount, datatype);
+   /* check input area's stated bounds make sense */
+   check_mem_is_addressable_untyped(inbuf, insize);
+   /* check input area's actual used size properly */
+   err = PMPI_Pack_size( outcount, datatype, comm, &szB );
+   if (err == MPI_SUCCESS && szB > 0) {
+      check_mem_is_addressable_untyped( 
+         ((char*)inbuf) + position_ORIG, szB
+      );
+   }
+
+   CALL_FN_W_7W(err, fn, inbuf,insize,position, outbuf,outcount,datatype, comm);
+
+   if (err == MPI_SUCCESS && (*position) > position_ORIG) {
+      /* recheck input more carefully */
+      check_mem_is_defined_untyped( 
+         ((char*)inbuf) + position_ORIG, *position - position_ORIG
+      );
+      /* paint output */
+      make_mem_defined_if_addressable( outbuf, outcount, datatype );
+   }
+   after("Unpack", err);
+   return err;
+}
+
+
 /*------------------------------------------------------------*/
 /*---                                                      ---*/
 /*--- Sec 4.4, Broadcast                                   ---*/
@@ -1829,7 +2077,7 @@ int WRAPPER_FOR(PMPI_Error_string)( int errorcode, char* string,
 
 /* --- Init --- */
 /* rd: *argc, *argv[0 .. *argc-1] */
-int WRAPPER_FOR(PMPI_Init)(int *argc, char ***argv)
+long WRAPPER_FOR(PMPI_Init)(int *argc, char ***argv)
 {
    OrigFn fn;
    int    err;
@@ -1839,7 +2087,10 @@ int WRAPPER_FOR(PMPI_Init)(int *argc, char ***argv)
    check_mem_is_defined_untyped(*argv, *argc * sizeof(char**));
    CALL_FN_W_WW(err, fn, argc,argv);
    after("Init", err);
-   return err;
+   if (opt_initkludge)
+      return (long)(void*)&mpiwrap_walk_type_EXTERNALLY_VISIBLE;
+   else
+      return (long)err;
 }
 
 /* --- Initialized --- */
@@ -2211,7 +2462,7 @@ DEFAULT_WRAPPER_W_1W(Op_f2c)
 DEFAULT_WRAPPER_W_1W(Op_free)
 DEFAULT_WRAPPER_W_7W(Pack_external)
 DEFAULT_WRAPPER_W_4W(Pack_external_size)
-DEFAULT_WRAPPER_W_7W(Pack)
+/* DEFAULT_WRAPPER_W_7W(Pack) */
 DEFAULT_WRAPPER_W_4W(Pack_size)
 /* int MPI_Pcontrol(const int level, ...) */
 /* DEFAULT_WRAPPER_W_4W(Probe) */
@@ -2287,7 +2538,7 @@ DEFAULT_WRAPPER_W_2W(Type_size)
 DEFAULT_WRAPPER_W_5W(Type_struct)
 DEFAULT_WRAPPER_W_2W(Type_ub)
 DEFAULT_WRAPPER_W_5W(Type_vector)
-DEFAULT_WRAPPER_W_7W(Unpack)
+/* DEFAULT_WRAPPER_W_7W(Unpack) */
 DEFAULT_WRAPPER_W_3W(Unpublish_name)
 DEFAULT_WRAPPER_W_7W(Unpack_external)
 /* DEFAULT_WRAPPER_W_3W(Waitall) */
