@@ -7,8 +7,8 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2005 Nicholas Nethercote <njn@valgrind.org>
-   Copyright (C) 2005 Cerion Armour-Brown <cerion@open-works.co.uk>
+   Copyright (C) 2005-2007 Nicholas Nethercote <njn@valgrind.org>
+   Copyright (C) 2005-2007 Cerion Armour-Brown <cerion@open-works.co.uk>
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -29,6 +29,8 @@
 */
 
 #include "pub_core_basics.h"
+#include "pub_core_vki.h"
+#include "pub_core_vkiscnums.h"
 #include "pub_core_threadstate.h"
 #include "pub_core_aspacemgr.h"
 #include "pub_core_debuglog.h"
@@ -50,8 +52,6 @@
 #include "priv_syswrap-linux.h"     /* for decls of linux-ish wrappers */
 #include "priv_syswrap-main.h"
 
-#include "vki_unistd.h"              /* for the __NR_* constants */
-
 
 /* ---------------------------------------------------------------------
    clone() handling
@@ -267,7 +267,7 @@ static SysRes do_clone ( ThreadId ptid,
    ThreadState* ctst = VG_(get_ThreadState)(ctid);
    ULong        word64;
    UWord*       stack;
-   NSegment*    seg;
+   NSegment const* seg;
    SysRes       res;
    vki_sigset_t blockall, savedmask;
 
@@ -341,6 +341,12 @@ static SysRes do_clone ( ThreadId ptid,
       ctst->client_stack_szB  = 0;
    }
 
+   /* Assume the clone will succeed, and tell any tool that wants to
+      know that this thread has come into existence.  If the clone
+      fails, we'll send out a ll_exit notification for it at the out:
+      label below, to clean up. */
+   VG_TRACK ( pre_thread_ll_create, ptid, ctid );
+
    if (flags & VKI_CLONE_SETTLS) {
       if (debug)
          VG_(printf)("clone child has SETTLS: tls at %p\n", child_tls);
@@ -374,6 +380,8 @@ static SysRes do_clone ( ThreadId ptid,
       /* clone failed */
       VG_(cleanup_thread)(&ctst->arch);
       ctst->status = VgTs_Empty;
+      /* oops.  Better tell the tool the thread exited in a hurry :-) */
+      VG_TRACK( pre_thread_ll_exit, ctid );
    }
 
    return res;
@@ -1000,60 +1008,13 @@ PRE(sys_clone)
    }
 }
 
-//zz PRE(sys_sigreturn)
-//zz {
-//zz    ThreadState* tst;
-//zz    PRINT("sigreturn ( )");
-//zz 
-//zz    vg_assert(VG_(is_valid_tid)(tid));
-//zz    vg_assert(tid >= 1 && tid < VG_N_THREADS);
-//zz    vg_assert(VG_(is_running_thread)(tid));
-//zz 
-//zz    ///* Adjust esp to point to start of frame; skip back up over
-//zz    //   sigreturn sequence's "popl %eax" and handler ret addr */
-//zz    tst = VG_(get_ThreadState)(tid);
-//zz    //tst->arch.vex.guest_ESP -= sizeof(Addr)+sizeof(Word);
-//zz    // Should we do something equivalent on ppc64?  Who knows.
-//zz 
-//zz    ///* This is only so that the EIP is (might be) useful to report if
-//zz    //   something goes wrong in the sigreturn */
-//zz    //ML_(fixup_guest_state_to_restart_syscall)(&tst->arch);
-//zz    // Should we do something equivalent on ppc64?  Who knows.
-//zz 
-//zz    VG_(sigframe_destroy)(tid, False);
-//zz 
-//zz    /* For unclear reasons, it appears we need the syscall to return
-//zz       without changing R3.  Since R3 is the return value, and can
-//zz       denote either success or failure, we must set up so that the
-//zz       driver logic copies it back unchanged.  Also, note R3 is of
-//zz       the guest registers written by VG_(sigframe_destroy). */
-//zz    /* jrs 16 Nov 05: for some reason this occasionally causes the 
-//zz       is-this-a-sane-error-value sanity check to fail:
-//zz       m_syswrap/syswrap-ppc64-linux.c:1037
-//zz         (vgSysWrap_ppc64_linux_sys_sigreturn_before): 
-//zz         Assertion 'wzz >= 0 && wzz < 10000' failed.
-//zz       Hence use a sanity-check-free version.  
-//zz       Perhaps we should ignore CR0.S0 here?
-//zz       In general I have no idea what this is for or if it is necessary.
-//zz       It's a conceptual copy-n-paste from the x86 equivalent, but I'm 
-//zz       equally unclear as to whether it is needed there either.
-//zz    */
-//zz    SET_STATUS_from_SysRes_NO_SANITY_CHECK(
-//zz       VG_(mk_SysRes_ppc64_linux)( 
-//zz          tst->arch.vex.guest_GPR3,
-//zz          /* get CR0.SO */
-//zz          (LibVEX_GuestPPC32_get_CR( &tst->arch.vex ) >> 28) & 1
-//zz       )
-//zz    );
-//zz 
-//zz    /* Check to see if some any signals arose as a result of this. */
-//zz    *flags |= SfPollAfter;
-//zz }
-
 PRE(sys_rt_sigreturn)
 {
+   /* See comments on PRE(sys_rt_sigreturn) in syswrap-amd64-linux.c for
+      an explanation of what follows. */
+
    ThreadState* tst;
-   PRINT("rt_sigreturn ( )");
+   PRINT("sys_rt_sigreturn ( )");
 
    vg_assert(VG_(is_valid_tid)(tid));
    vg_assert(tid >= 1 && tid < VG_N_THREADS);
@@ -1063,25 +1024,22 @@ PRE(sys_rt_sigreturn)
    //   ret addr */
    tst = VG_(get_ThreadState)(tid);
    //tst->arch.vex.guest_ESP -= sizeof(Addr);
-   // Should we do something equivalent on ppc64?  Who knows.
+   // Should we do something equivalent on ppc64-linux?  Who knows.
 
    ///* This is only so that the EIP is (might be) useful to report if
    //   something goes wrong in the sigreturn */
    //ML_(fixup_guest_state_to_restart_syscall)(&tst->arch);
    // Should we do something equivalent on ppc64?  Who knows.
 
+   /* Restore register state from frame and remove it */
    VG_(sigframe_destroy)(tid, True);
 
-   /* See comments above in PRE(sys_sigreturn) about this. */
-   SET_STATUS_from_SysRes_NO_SANITY_CHECK(
-      VG_(mk_SysRes_ppc64_linux)( 
-         tst->arch.vex.guest_GPR3,
-         /* get CR0.SO */
-         (LibVEX_GuestPPC64_get_CR( &tst->arch.vex ) >> 28) & 1
-      )
-   );
+   /* Tell the driver not to update the guest state with the "result",
+      and set a bogus result to keep it happy. */
+   *flags |= SfNoWriteResult;
+   SET_STATUS_Success(0);
 
-   /* Check to see if some any signals arose as a result of this. */
+   /* Check to see if any signals arose as a result of this. */
    *flags |= SfPollAfter;
 }
 
@@ -1298,7 +1256,7 @@ const SyscallTableEntry ML_(syscall_table)[] = {
 // _____(__NR_syslog,            sys_syslog),             // 103
    GENXY(__NR_setitimer,         sys_setitimer),          // 104
 
-// _____(__NR_getitimer,         sys_getitimer),          // 105
+   GENXY(__NR_getitimer,         sys_getitimer),          // 105
    GENXY(__NR_stat,              sys_newstat),            // 106
    GENXY(__NR_lstat,             sys_newlstat),           // 107
    GENXY(__NR_fstat,             sys_newfstat),           // 108
@@ -1492,10 +1450,10 @@ const SyscallTableEntry ML_(syscall_table)[] = {
    LINX_(__NR_mq_notify,         sys_mq_notify),          // 266
    LINXY(__NR_mq_getsetattr,     sys_mq_getsetattr),      // 267
 // _____(__NR_kexec_load,        sys_kexec_load),         // 268
-// _____(__NR_add_key,           sys_add_key),            // 269
+   LINX_(__NR_add_key,           sys_add_key),            // 269
 
-// _____(__NR_request_key,       sys_request_key),        // 270
-// _____(__NR_keyctl,            sys_keyctl),             // 271
+   LINX_(__NR_request_key,       sys_request_key),        // 270
+   LINXY(__NR_keyctl,            sys_keyctl),             // 271
 // _____(__NR_waitid,            sys_waitid),             // 272
 // _____(__NR_ioprio_set,        sys_ioprio_set),         // 273
 // _____(__NR_ioprio_get,        sys_ioprio_get),         // 274
