@@ -40,7 +40,7 @@ VgToolInterface VG_(tdict);
 
 void VG_(basic_tool_funcs)(
    void(*post_clo_init)(void),
-   IRBB*(*instrument)(VgCallbackClosure*, IRBB*, 
+   IRSB*(*instrument)(VgCallbackClosure*, IRSB*, 
                       VexGuestLayout*, VexGuestExtents*, IRType, IRType),
    void(*fini)(Int)
 )
@@ -86,7 +86,7 @@ VgNeeds VG_(needs) = {
    .core_errors          = False,
    .tool_errors          = False,
    .libc_freeres         = False,
-   .basic_block_discards = False,
+   .superblock_discards  = False,
    .command_line_options = False,
    .client_requests      = False,
    .syscall_wrapper      = False,
@@ -94,6 +94,7 @@ VgNeeds VG_(needs) = {
    .data_syms	         = False,
    .malloc_replacement   = False,
    .xml_output           = False,
+   .final_IR_tidy_pass   = False
 };
 
 /* static */
@@ -164,17 +165,18 @@ NEEDS(core_errors)
 NEEDS(data_syms)
 NEEDS(xml_output)
 
-void VG_(needs_basic_block_discards)(
+void VG_(needs_superblock_discards)(
    void (*discard)(Addr64, VexGuestExtents)
 )
 {
-   VG_(needs).basic_block_discards = True;
-   VG_(tdict).tool_discard_basic_block_info = discard;
+   VG_(needs).superblock_discards = True;
+   VG_(tdict).tool_discard_superblock_info = discard;
 }
 
 void VG_(needs_tool_errors)(
    Bool (*eq)         (VgRes, Error*, Error*),
    void (*pp)         (Error*),
+   Bool show_TIDs,
    UInt (*update)     (Error*),
    Bool (*recog)      (Char*, Supp*),
    Bool (*read_extra) (Int, Char*, Int, Supp*),
@@ -186,6 +188,7 @@ void VG_(needs_tool_errors)(
    VG_(needs).tool_errors = True;
    VG_(tdict).tool_eq_Error                     = eq;
    VG_(tdict).tool_pp_Error                     = pp;
+   VG_(tdict).tool_show_ThreadIDs_for_errors    = show_TIDs;
    VG_(tdict).tool_update_extra                 = update;
    VG_(tdict).tool_recognised_suppression       = recog;
    VG_(tdict).tool_read_extra_suppression_info  = read_extra;
@@ -260,6 +263,13 @@ void VG_(needs_malloc_replacement)(
    VG_(tdict).tool_client_redzone_szB   = client_malloc_redzone_szB;
 }
 
+void VG_(needs_final_IR_tidy_pass)( 
+   IRSB*(*final_tidy)(IRSB*)
+)
+{
+   VG_(needs).final_IR_tidy_pass = True;
+   VG_(tdict).tool_final_IR_tidy_pass = final_tidy;
+}
 
 /*--------------------------------------------------------------------*/
 /* Tracked events */
@@ -321,14 +331,12 @@ DEF(track_post_reg_write,        CorePart, ThreadId,        OffT, SizeT)
 
 DEF(track_post_reg_write_clientcall_return, ThreadId, OffT, SizeT, Addr)
 
-DEF(track_thread_run,            ThreadId)
+DEF(track_start_client_code,     ThreadId, ULong)
+DEF(track_stop_client_code,      ThreadId, ULong)
 
-DEF(track_post_thread_create,    ThreadId, ThreadId)
-DEF(track_post_thread_join,      ThreadId, ThreadId)
-
-DEF(track_pre_mutex_lock,        ThreadId, void*)
-DEF(track_post_mutex_lock,       ThreadId, void*)
-DEF(track_post_mutex_unlock,     ThreadId, void*)
+DEF(track_pre_thread_ll_create,  ThreadId, ThreadId)
+DEF(track_pre_thread_first_insn, ThreadId)
+DEF(track_pre_thread_ll_exit,    ThreadId)
 
 DEF(track_pre_deliver_signal,    ThreadId, Int sigNo, Bool)
 DEF(track_post_deliver_signal,   ThreadId, Int sigNo)
