@@ -61,6 +61,12 @@ extern Int   VG_(clo_sanity_level);
 extern Bool  VG_(clo_demangle);
 /* Simulate child processes? default: NO */
 extern Bool  VG_(clo_trace_children);
+/* After a fork, the child's output can become confusingly
+   intermingled with the parent's output.  This is especially
+   problematic when VG_(clo_xml) is True.  Setting
+   VG_(clo_child_silent_after_fork) causes children to fall silent
+   after fork() calls. */
+extern Bool  VG_(clo_child_silent_after_fork);
 
 /* Where logging output is to be sent to.
 
@@ -68,16 +74,11 @@ extern Bool  VG_(clo_trace_children);
    taken from the command line.  (fd 2, stderr, is the default.)
    clo_log_name is irrelevant.
 
-   With --log-file/--log-file-exactly, clo_log_name holds the log-file
-   name, and is taken from the command line.  clo_log_fd is then
-   made to hold the relevant file id, by opening clo_log_name
-   (concatenated with the process ID) for writing.
-
-   With --log-file, there is an additional twist: if
-   clo_log_file_qualifier is non-NULL, the contents of the environment
-   variable specified by clo_log_file_qualifier is incorporated into
-   the logfile name.  This is useful in that it allows the logfile
-   name to incorporate environmental information.
+   With --log-file, clo_log_name holds the log-file name, and is taken from
+   the command line (and possibly has process ID/env var contents in it, if
+   the %p or %q format specifiers are used).  clo_log_fd is then made to
+   hold the relevant file id, by opening clo_log_name (concatenated with the
+   process ID) for writing.
 
    With --log-socket, clo_log_name holds the hostname:portnumber pair,
    and is taken from the command line.  clo_log_fd is then made to hold
@@ -88,7 +89,6 @@ extern Bool  VG_(clo_trace_children);
    (stderr). */
 extern Int   VG_(clo_log_fd);
 extern Char* VG_(clo_log_name);
-extern Char* VG_(clo_log_file_qualifier);
 
 /* Add timestamps to log messages?  default: NO */
 extern Bool  VG_(clo_time_stamp);
@@ -101,9 +101,9 @@ extern Int   VG_(clo_n_suppressions);
 extern Char* VG_(clo_suppressions)[VG_CLO_MAX_SFILES];
 
 /* DEBUG: print generated code?  default: 00000000 ( == NO ) */
-extern Bool  VG_(clo_trace_flags);
+extern UChar VG_(clo_trace_flags);
 /* DEBUG: do bb profiling?  default: 00000000 ( == NO ) */
-extern Bool  VG_(clo_profile_flags);
+extern UChar VG_(clo_profile_flags);
 /* DEBUG: if tracing codegen, be quiet until after this bb ( 0 ) */
 extern Int   VG_(clo_trace_notbelow);
 /* DEBUG: print system calls?  default: NO */
@@ -112,8 +112,16 @@ extern Bool  VG_(clo_trace_syscalls);
 extern Bool  VG_(clo_trace_signals);
 /* DEBUG: print symtab details?  default: NO */
 extern Bool  VG_(clo_trace_symtab);
+/* DEBUG: restrict symtab etc details to object name pattern.  Default: "*" */
+extern HChar* VG_(clo_trace_symtab_patt);
 /* DEBUG: print call-frame-info details?  default: NO */
 extern Bool  VG_(clo_trace_cfi);
+/* DEBUG:  mimic /usr/bin/readelf --syms?  default: NO */
+extern Bool  VG_(clo_debug_dump_syms);
+/* DEBUG: mimic /usr/bin/readelf --debug-dump=line?  default: NO */
+extern Bool  VG_(clo_debug_dump_line);
+/* DEBUG: mimic  /usr/bin/readelf --debug-dump=frames?  default: NO */
+extern Bool  VG_(clo_debug_dump_frames);
 /* DEBUG: print redirection details?  default: NO */
 extern Bool  VG_(clo_trace_redir);
 /* DEBUG: print thread scheduling events?  default: NO */
@@ -127,6 +135,8 @@ extern Int   VG_(clo_dump_error);
 extern Int   VG_(clo_backtrace_size);
 /* Engage miscellaneous weird hacks needed for some progs. */
 extern Char* VG_(clo_sim_hints);
+/* Show symbols in the form 'name+offset' ?  Default: NO */
+extern Bool VG_(clo_sym_offsets);
 
 /* Track open file descriptors? */
 extern Bool  VG_(clo_track_fds);
@@ -139,15 +149,14 @@ extern Bool  VG_(clo_track_fds);
 extern Bool  VG_(clo_run_libc_freeres);
 /* Continue stack traces below main()?  Default: NO */
 extern Bool VG_(clo_show_below_main);
-/* Model the pthread library */
-extern Bool VG_(clo_model_pthreads);
 
 /* Should we show VEX emulation warnings?  Default: NO */
 extern Bool VG_(clo_show_emwarns);
 
 /* How much does the stack pointer have to change before tools
-   consider a stack switch to have happened?  Default: 2000000 bytes */
-extern Int VG_(clo_max_stackframe);
+   consider a stack switch to have happened?  Default: 2000000 bytes
+   NB: must be host-word-sized to be correct (hence Word). */
+extern Word VG_(clo_max_stackframe);
 
 /* Delay startup to allow GDB to be attached?  Default: NO */
 extern Bool VG_(clo_wait_for_gdb);
@@ -172,6 +181,19 @@ extern VgSmc VG_(clo_smc_check);
    so they can be properly handled by m_syswrap. */
 extern HChar* VG_(clo_kernel_variant);
 
+/* --------- Functions --------- */
+
+/* Call this if the executable is missing.  This function prints an
+   error message, then shuts down the entire system. */
+__attribute__((noreturn))
+extern void VG_(err_missing_prog) ( void );
+
+/* Similarly - complain and stop if there is some kind of config
+   error. */
+__attribute__((noreturn))
+extern void VG_(err_config_error) ( Char* msg );
+
+
 #endif   // __PUB_CORE_OPTIONS_H
 
 /*--------------------------------------------------------------------*/
