@@ -134,7 +134,7 @@ extern Int VG_(am_get_segment_starts)( Addr* starts, Int nStarts );
 
 
 // See pub_core_aspacemgr.h for description.
-extern NSegment* VG_(am_find_nsegment) ( Addr a ); 
+extern NSegment const * VG_(am_find_nsegment) ( Addr a ); 
 
 // See pub_core_aspacemgr.h for description.
 extern HChar* VG_(am_get_filename)( NSegment* );
